
/* Sxgen  rom and microprogram generator for sx chip

   sx micro						10 July 2006

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University
*/

#include "sx.h"

char ms[MWIDTH]; 			// one word, micro store
sym mads[100];				// symbol for microads
int udop[MAXOP]; 			// micro rom
int mx[MLEN][MWIDTH];		// micro event-list
int mw[MLEN+1];				// point to begin of each event-list
int nxt[MLEN];				// next micro address

FILE *fo;

// read microprogram mpgm.txt
void readmpgm(void){
  int i,n;
  FILE *fp;
  char ibuf[1000];

  mw[0] = 0;
  fp = fopen("mpgm.txt","r");
  fgets(ibuf,1000,fp);
  sscanf(ibuf,"%d",&n);	  /* length of mpgm */
  for( i=0; i<n; i++ ) {
    fgets(ibuf,1000,fp);
    strncpy(ms,ibuf+4,MWIDTH);  // be careful ibuf+4
	scan1(i);
  }
  fclose(fp);
}

// read mspec.h to get microaddress
void readmspec(void){
	int i,n,v;
	FILE *fp;
	char w1[80],w2[80];

	fp = fopen("mspec.h","r");
	i = 0;
	while(1){
		fscanf(fp,"%s%s%d",w1,w2,&v);
		fprintf(fo,"%s %s %d\n",w1,w2,v);
		if( eqs(w2,"MLEN") ){	// last symbol
			break;
		}else if( w2[0] == 'a' ){
			strcpy(mads[i].name,w2+2);
			mads[i].val = v;
			i++;
		}
	}
	mads[i].name[0] = 0;			// end of symbol
}

/* test current ms is set at pos */
int set(int pos ){
	return( ms[pos] == '1' );
}

// lookup name-val pair in mads, 0 not found
int lookup(char *s){
	int i;
	i = 0;
	while(mads[i].name[0] != 0){
		if( eqs(mads[i].name,s) ) return mads[i].val;
		i++;
	}
	printf("not found %s in mads\n",s);
	return 0;
}

// put micro-address into rom : udop[]
void initrom(void){
	udop[icAdd] = lookup("bop");
	udop[icSub] = lookup("bop");
	udop[icMul] = lookup("bop");
	udop[icDiv] = lookup("bop");
	udop[icBand] = lookup("bop");
	udop[icBor] = lookup("bop");
	udop[icBxor] = lookup("bop");
	udop[icNot] = lookup("uop");
	udop[icEq] = lookup("bop");
	udop[icNe] = lookup("bop");
	udop[icLt] = lookup("bop");
	udop[icLe] = lookup("bop");
	udop[icGt] = lookup("bop");
	udop[icGe] = lookup("bop");
	udop[icShl] = lookup("bop");
	udop[icShr] = lookup("bop");
	udop[icLdx] = lookup("ldx");
	udop[icStx] = lookup("stx");
	udop[icRet] = lookup("ret");
	udop[icArray] = lookup("array");
	udop[icEnd] = lookup("end");
	udop[icGet] = lookup("get");
	udop[icPut] = lookup("put");
	udop[icLd] = lookup("ld");
	udop[icSt] = lookup("st");
	udop[icLit] = lookup("lit");
	udop[icJmp] = lookup("jmp");
	udop[icJt] = lookup("jt");
	udop[icJf] = lookup("jf");
	udop[icCall] = lookup("call");
	udop[icInc] = lookup("inc");
	udop[icDec] = lookup("dec");
	udop[icSys] = lookup("sys");
	udop[icDjnz] = lookup("djnz");  // new instruction
}

int ugoto(void){	/* decode Goto field */
	int i, a, w;
	a = 0;
	w = 1;
	for(i=MWIDTH-1; i>=MCWIDTH; i--){
		if( set(i) )   a += w;
		w += w;
	}
	return a;
}

// order of events on data path
int sig[] = {
	s_x_ts, s_x_fp, s_x_sp, s_x_nx, s_y_ff, s_y_arg,
	s_alu_add, s_alu_sub, s_alu_inc, s_alu_dec, s_alu_z,
	s_alu_eq, s_alu_p1, s_alu_p2, s_alu_op,
	s_a_pc, s_a_tbus, s_d_fp, s_d_ts,
	s_mR, s_mW, s_b_dbus, s_b_tbus, s_b_pc,
	s_j_pc1, s_j_pcarg, s_j_tbus,
	s_lpc, s_lir, s_lts, s_lfp, s_lsp, s_lnx, s_lff,
	s_ifT, s_ifF, s_decode, s_trap, -1
};

// convert one word of ms to mx
void scan1(int a){
	int i, k, s;
	i = 0;
	k = 0;
	while( sig[i] >= 0 ){
		if( set(sig[i]) ){
		   mx[a][k] = sig[i];
		   k++;
		}
		i++;
	}
	mx[a][k] = -1;	// each word is terminated by -1
	mw[a+1] = mw[a] + k + 1; // beginning of next word
	nxt[a] = ugoto();
}

void outbit(void){
	int i, k;
	// udop[]
	fprintf(fo,"\nint udop[] = {");
	for(i=0; i<MAXOP; i++){
		if(i%10 == 0) fprintf(fo,"\n");
		fprintf(fo,"%d, ",udop[i]);
	}
	fprintf(fo,"0 };\n");
	// mw[]
	fprintf(fo,"\nint mw[] = {");
	for(i=0; i<MLEN; i++){
		if( i%10 == 0 ) fprintf(fo,"\n");
		fprintf(fo,"%d, ",mw[i]);
	}
	fprintf(fo,"0 };\n");
	// mx[]
	fprintf(fo,"\nint mx[] = {\n");
	for(i=0; i<MLEN; i++){
		k = 0;
		while(mx[i][k] >= 0){
			fprintf(fo,"%d, ",mx[i][k]);
			k++;
		}
		fprintf(fo,"-1,\n");
	}
	fprintf(fo,"0 };\n");
	// nxt[]
	fprintf(fo,"\nint nxt[] = {");
	for(i=0; i<MLEN; i++){
		if( i%10 == 0 ) fprintf(fo,"\n");
		fprintf(fo,"%d, ",nxt[i]);
	}
	fprintf(fo,"0 };\n");
}

int main(void){
	fo = fopen("sxbit.h","w");
	readmspec();
	readmpgm();
	initrom();
	outbit();
	fclose(fo);
	return 0;
}

